from executor import run_calculate_function_value


def func(a, b, *, c , d):
    return a + b + c + d


class A:
    def __init__(self, x):
        self.x = x

    def func(self, a, b, *, c, d):
        return a + b + c + d + self.x


def test_run_function():
    run_calculate_function_value(
        'database_func',
        func,
        [1, 2],
        {'c': 3, 'd': 4},
        'tests.py',
        'test_output_1'
    )
    with open('test_output_1', 'r') as out:
        actual = [line.strip() for line in out.readlines()]

    expected = [
        'success',
        '{"json": {"type": "builtins.int", "value": "10", "strategy": "repr", "comparable": true}, "memory": {}}',
        '[4, 5]',
        '[]'
    ]

    assert actual == expected


def test_run_method():
    a = A(100)
    run_calculate_function_value(
        'database_func_A',
        a.func,
        [1, 2],
        {'c': 3, 'd': 4},
        'tests.py',
        'test_output_2'
    )
    with open('test_output_2', 'r') as out:
        actual = [line.strip() for line in out.readlines()]

    expected = [
        'success',
        '{"json": {"type": "builtins.int", "value": "110", "strategy": "repr", "comparable": true}, "memory": {}}',
        '[12, 13]',
        '[]'
    ]

    assert actual == expected



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['utbot_executor']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=6.5.0,<7.0.0', 'deep-serialization>=0.1.11,<0.2.0']

entry_points = \
{'console_scripts': ['utbot-executor = utbot_executor:utbot_executor']}

setup_kwargs = {
    'name': 'utbot-executor',
    'version': '0.2.7',
    'description': '',
    'long_description': '# UtBot Executor\n\nUtil for python code execution and state serialization.\n',
    'author': 'Vyacheslav Tamarin',
    'author_email': 'vyacheslav.tamarin@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Python code executor for UnitTestBot"""
from typing import Any, Callable, Dict, List, Optional, Tuple
import json
import inspect
import coverage

from deep_serialization.deep_serialization import serialize_objects


from utbot_executor.serializer import PythonTreeSerializer
from utbot_executor.utils import suppress_stdout


def __get_lines(start, end, lines):
    return [x for x in lines if start < x < end]


def serialize_state(
        args: List[Any],
        kwargs: Dict[str, Any],
        result: Any = None,
        ) -> Tuple[List[str], List[str], str, str]:
    all_arguments = args + list(kwargs.values()) + [result]
    ids, serialized_memory = serialize_objects(all_arguments)
    return (
            ids[:len(args)],
            ids[len(args):len(args)+len(kwargs)],
            ids[-1],
            serialized_memory,
            )


def run_calculate_function_value(
        database_name: str,
        function: Callable,
        args: List[Any],
        kwargs: Dict[str, Any],
        fullpath: str,
        output: str
    ):
    _, _, _, state_before = serialize_state(args, kwargs)

    __is_exception = False
    __cov = coverage.Coverage(
        data_file=database_name,
        data_suffix=".coverage",
    )
    __cov.start()
    try:
        with suppress_stdout():
            __result = function(*args, **kwargs)
    except Exception as __exception:
        __result = __exception
        __is_exception = True
    __cov.stop()

    (__sources, __start, ) = inspect.getsourcelines(function)
    __end = __start + len(__sources)
    (_, __stmts, _, __missed, _, ) = __cov.analysis2(fullpath)
    __cov.erase()
    __stmts_filtered = __get_lines(__start, __end, __stmts)
    __stmts_filtered_with_def = [__start] + __stmts_filtered
    __missed_filtered = __get_lines(__start, __end, __missed)

    args_ids, kwargs_ids, result_id, state_after = serialize_state(args, kwargs, __result)

    with open(output, "w", encoding="utf-8") as __out_file:
        __output_data = json.dumps({
            "status": "success",
            "isException": __is_exception,
            "statements": __stmts_filtered_with_def,
            "missedStatements": __missed_filtered,
            "stateBefore": state_before,
            "stateAfter": state_after,
            "argsIds": args_ids,
            "kwargsIds": kwargs_ids,
            "resultId": result_id,
        })
        __out_file.write(__output_data)


def fail_argument_initialization(output: str, exception: Exception):
    with open(output, "w", encoding="utf-8") as __out_file:
        __output_data = json.dumps({
            "status": "fail",
            'exception': exception,
        })
        __out_file.write(exception)



import pickle
from itertools import zip_longest


class PythonTreeSerializer:
    class MemoryObj:
        def __init__(self, json):
            self.json = json
            self.deserialized_obj = None
            self.comparable = False
            self.is_draft = True

    def __init__(self):
        self.memory = {}

    def memory_view(self):
        return ' | '.join(f'{id_}: {obj.deserialized_obj}' for id_, obj in self.memory.items())

    @staticmethod
    def get_type(py_object):
        if py_object is None:
            return 'types.NoneType'
        module = type(py_object).__module__
        return '{module}.{name}'.format(
            module=module,
            name=type(py_object).__name__,
        )

    @staticmethod
    def get_type_name(type_):
        if type_ is None:
            return 'types.NoneType'
        return '{module}.{name}'.format(
            module=type_.__module__,
            name=type_.__name__,
        )

    @staticmethod
    def has_reduce(py_object) -> bool:
        if getattr(py_object, '__reduce__', None) is None:
            return False
        else:
            try:
                py_object.__reduce__()
                return True
            except TypeError:
                return False

    def save_to_memory(self, id_, py_json, deserialized_obj):
        mem_obj = PythonTreeSerializer.MemoryObj(py_json)
        mem_obj.deserialized_obj = deserialized_obj
        self.memory[id_] = mem_obj
        return mem_obj

    def get_reduce(self, py_object):
        id_ = id(py_object)

        py_object_reduce = py_object.__reduce__()
        reduce_value = [
            default if obj is None else obj
            for obj, default in zip_longest(
                py_object_reduce,
                [None, [], {}, [], []],
                fillvalue=None
            )
        ]

        constructor = PythonTreeSerializer.get_type_name(reduce_value[0])
        args, deserialized_args = PythonTreeSerializer.unzip_list([
            self.serialize(arg)
            for arg in reduce_value[1]
        ])
        json_obj = {
            'id': id_,
            'type': PythonTreeSerializer.get_type(py_object),
            'constructor': constructor,
            'args': args,
            'state': [],
            'listitems': [],
            'dictitems': [],
        }
        deserialized_obj = reduce_value[0](*deserialized_args)
        memory_obj = self.save_to_memory(id_, json_obj, deserialized_obj)

        state, deserialized_state = self.unzip_dict([
            (attr, self.serialize(value))
            for attr, value in reduce_value[2].items()
        ], skip_first=True)
        listitems, deserialized_listitems = self.unzip_list([
            self.serialize(item)
            for item in reduce_value[3]
        ])
        dictitems, deserialized_dictitems = self.unzip_dict([
            (self.serialize(key), self.serialize(value))
            for key, value in reduce_value[4]
        ])

        memory_obj.json['state'] = state
        memory_obj.json['listitems'] = listitems
        memory_obj.json['dictitems'] = dictitems

        for key, value in deserialized_state.items():
            setattr(deserialized_obj, key, value)
        for item in deserialized_listitems:
            deserialized_obj.append(item)
        for key, value in deserialized_dictitems.items():
            deserialized_obj[key] = value

        memory_obj.deserialized_obj = deserialized_obj
        memory_obj.is_draft = False

        return id_, deserialized_obj

    def serialize(self, py_object):
        type_ = PythonTreeSerializer.get_type(py_object)
        id_ = id(py_object)
        skip_comparable = False
        comparable = True

        if id_ in self.memory:
            value = id_
            strategy = 'memory'
            skip_comparable = True
            comparable = False
            deserialized_obj = self.memory[id_].deserialized_obj
            if not self.memory[id_].is_draft:
                self.memory[id_].comparable = py_object == deserialized_obj
                skip_comparable = False
        elif isinstance(py_object, type):
            value = PythonTreeSerializer.get_type_name(py_object)
            strategy = 'repr'
            deserialized_obj = py_object
        elif any(type(py_object) == t for t in (list, set, tuple)):
            elements = [
                self.serialize(element) for element in py_object
            ]
            value, deserialized_obj = PythonTreeSerializer.unzip_list(elements, type(py_object))
            comparable = all([element['comparable'] for element in value])
            strategy = 'generic'
        elif type(py_object) == dict:
            elements = [
                [self.serialize(key), self.serialize(value)]
                for key, value in py_object.items()
            ]
            value, deserialized_obj = PythonTreeSerializer.unzip_dict(elements)
            comparable = all([element[1]['comparable'] for element in value])
            strategy = 'generic'
        elif PythonTreeSerializer.has_reduce(py_object):
            value, deserialized_obj = self.get_reduce(py_object)
            strategy = 'memory'
        else:
            value = repr(py_object)
            try:
                deserialized_obj = pickle.loads(pickle.dumps(py_object))
            except Exception:
                deserialized_obj = py_object
                skip_comparable = True
                comparable = False
            strategy = 'repr'

        if not skip_comparable:
            try:
                comparable = comparable and (py_object == deserialized_obj)
            except Exception:
                comparable = False

        return {
                   'type': type_,
                   'value': value,
                   'strategy': strategy,
                   'comparable': comparable,
               }, deserialized_obj

    @staticmethod
    def unzip_list(elements, cast_second=list):
        if len(elements) == 0:
            first, second = [], []
        else:
            first, second = list(zip(*elements))
        return first, cast_second(second)

    @staticmethod
    def unzip_dict(elements, cast_second=dict, skip_first=False):
        if len(elements) == 0:
            first, second = [], []
        else:
            if skip_first:
                first = [[element[0], element[1][0]] for element in elements]
                second = [[element[0], element[1][1]] for element in elements]
            else:
                first = [[element[0][0], element[1][0]] for element in elements]
                second = [[element[0][1], element[1][1]] for element in elements]
        return first, cast_second(second)

    def dumps(self, obj):
        return {
            'json': self.serialize(obj)[0],
            'memory': {
                key: value.json
                for key, value in self.memory.items()
            }
        }



import os
import sys
from contextlib import contextmanager


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
